# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sys
import tempfile
import shutil
import logging
import pkg_resources


# NOTE: by inserting cwd (which *should* be the top-level source code
# dir, with 'ferenda' and 'test' as subdirs) into sys.path as early as
# possible, we make it possible for pkg_resources to find resources in
# the 'ferenda' package even when we change the cwd later on. We also
# have to call a resource method to make it stick.
sys.path.insert(0,os.getcwd())
pkg_resources.resource_listdir('ferenda','res')

from ferenda.manager import setup_logger; setup_logger('CRITICAL')
from ferenda.compat import unittest, OrderedDict, Mock, MagicMock, patch, call
from ferenda.testutil import RepoTester, FerendaTestCase

import six
from six.moves import configparser, reload_module
builtins = "__builtin__" if six.PY2 else "builtins"

from lxml import etree as ET
import requests.exceptions

from ferenda import manager, decorators, util, errors
from ferenda import DocumentRepository, LayeredConfig, DocumentStore

class staticmockstore(DocumentStore):
    def list_basefiles_for(cls,action):
        return ["arg1","myarg","arg2"]

class staticmockclass(DocumentRepository):
    """Example class for testing"""
    alias = "staticmock"
    resourcebase = None
    documentstore_class = staticmockstore
    @decorators.action
    def mymethod(self, arg):
        """Frobnicate the bizbaz"""
        if arg == "myarg":
            return "ok!"

    def download(self):
        return "%s download ok" % self.alias

    def parse(self, basefile):
        return "%s parse %s" % (self.alias, basefile)

    def relate(self, basefile):
        return "%s relate %s" % (self.alias, basefile)

    def generate(self, basefile): 
        return "%s generate %s" % (self.alias, basefile)

    def toc(self): 
        return "%s toc ok" % (self.alias)

    def news(self): 
        return "%s news ok" % (self.alias)

    def internalmethod(self, arg):
        pass

    @classmethod
    def setup(cls, action, config): pass
    @classmethod
    def teardown(cls, action, config): pass
        
        
    def get_default_options(self):
        opts = super(staticmockclass, self).get_default_options()
        opts.update({'datadir': 'data',
                     'loglevel': 'DEBUG',
                     'cssfiles': [self.resourcebase + '/test.css'],
                     'jsfiles': [self.resourcebase + '/test.js']})
        return opts
                    
    
    
class staticmockclass2(staticmockclass):
    """Another class for testing"""
    alias="staticmock2"
    def mymethod(self, arg):
        """Frobnicate the bizbaz (alternate implementation)"""
        if arg == "myarg":
            return "yeah!"

class staticmockclass3(staticmockclass):
    """Yet another (overrides footer())"""
    alias="staticmock3"
    def footer(self):
        return (("About", "http://example.org/about"),
                ("Legal", "http://example.org/legal"),
                ("Contact", "http://example.org/contact")
        )
    
class API(unittest.TestCase, FerendaTestCase):
    """Test cases for API level methods of the manager modules (functions
       like enable and makeresources, including unit tests of internal
       helpers.

    """
    def setUp(self):
        self.maxDiff = None
        self.tempdir = tempfile.mkdtemp()
        staticmockclass.resourcebase = self.tempdir
        # FIXME: this creates (and tearDown deletes) a file in
        # cwd. Should be placed in self.tempdir, but tests need to be
        # adjusted to find it there.

        # NB: The section keys are different from the specified
        # classes alias properties. This is intended.
        util.writefile("ferenda.ini", """[__root__]
datadir = %s
loglevel = CRITICAL
[test]
class=testManager.staticmockclass
[test2]
class=testManager.staticmockclass2
"""%self.tempdir)
        util.writefile(self.tempdir+"/test.js", "// test.js code goes here")
        util.writefile(self.tempdir+"/test.css", "/* test.css code goes here */")
        util.writefile(self.tempdir+"/transformed.scss", "a { color: red + green; }")

    def tearDown(self):
        if os.path.exists("ferenda.ini"):
            os.remove("ferenda.ini")
        shutil.rmtree(self.tempdir)
        
    def test_filter_argv(self):
        self.assertEqual(manager._filter_argv(["ecj", "parse", "62008J0034", "62008J0035"]),
                         ("ecj", "parse", ["62008J0034", "62008J0035"]))
        self.assertEqual(manager._filter_argv(["ecj", "parse", "62008J0034", "--force=True", "--frobnicate"]),
                         ("ecj", "parse", ["62008J0034"]))
        self.assertEqual(manager._filter_argv(["ecj", "--frobnicate"]),
                         ("ecj", None, []))

    def test_enable_class(self):
        # 1. test that a normal enabling goes well
        manager.enable("testManager.staticmockclass")
        # os.system("cat ferenda.ini")
        cfg = configparser.ConfigParser()
        cfg.read(["ferenda.ini"])
        self.assertEqual(cfg.get("staticmock","class"), "testManager.staticmockclass")
        # 2. test that an attempt to enable a nonexistent class fails"
        with self.assertRaises(ImportError):
            manager.enable("testManager.Nonexistent")

        # 3. test that an attempt to enable an alias fails
        with self.assertRaises(ValueError):
            manager.enable("staticmock")

    def test_run_class(self):
        enabled_classes = {'test':'testManager.staticmockclass'}
        argv = ["test", "mymethod","myarg"]
        self.assertEqual(manager._run_class(enabled_classes,argv),"ok!")

                    
    def test_list_enabled_classes(self):
        self.assertEqual(manager._list_enabled_classes(),
                         OrderedDict((("test","Example class for testing"),
                                      ("test2","Another class for testing"))))

    def test_list_class_usage(self):
        self.assertEqual(manager._list_class_usage(staticmockclass),
                         {'mymethod':'Frobnicate the bizbaz'})


    def test_makeresources(self):
        # Test1: No combining, resources specified by docrepos
        test = staticmockclass()
        # print("test.get_default_options %r" % test.get_default_options())
        test2 = staticmockclass2()
        s = os.sep
        want = {'css':[s.join(['rsrc', 'css','test.css'])],
                'js':[s.join(['rsrc', 'js','test.js'])],
                'xml':[s.join(['rsrc', 'resources.xml'])]
        }
        got = manager.makeresources([test,test2],self.tempdir+os.sep+'rsrc')
        self.assertEqual(want, got)
        tree = ET.parse(self.tempdir+os.sep+got['xml'][0])
        stylesheets=tree.find("stylesheets").getchildren()
        self.assertEqual(len(stylesheets),1)
        self.assertEqual(stylesheets[0].attrib['href'],'rsrc/css/test.css')
        javascripts=tree.find("javascripts").getchildren()
        self.assertEqual(len(javascripts),1)
        self.assertEqual(javascripts[0].attrib['src'],'rsrc/js/test.js')
        self.assertEqual(tree.find("sitename").text,"MySite")
        self.assertEqual(tree.find("sitedescription").text,"Just another Ferenda site")
        self.assertEqual(tree.find("url").text,"http://localhost:8000/")
        self.assertTrue(os.path.exists(self.tempdir+'/rsrc/css/test.css'))
        self.assertTrue(os.path.exists(self.tempdir+'/rsrc/js/test.js'))
        tabs=tree.find("tabs")
        self.assertTrue(tabs is not None)
        search=tree.find("search")
        self.assertTrue(search is not None)

        # Test2: combining, resources specified by global config
        # (maybe we should use smaller CSS+JS files? Test takes 2+ seconds...)
        want = {'css':[s.join(['rsrc', 'css','combined.css'])],
                'js':[s.join(['rsrc', 'js','combined.js'])],
                'xml':[s.join(['rsrc', 'resources.xml'])]
        }
        got = manager.makeresources([test,test2],self.tempdir+os.sep+'rsrc',
                                    combine=True,
                                    cssfiles=['res/css/normalize-1.1.3.css',
                                              'res/css/main.css'],
                                    jsfiles=['res/js/jquery-1.10.2.js',
                                             'res/js/modernizr-2.6.3.js',
                                             'res/js/respond-1.3.0.js'],
                                    sitename="Blahonga",
                                    sitedescription="A non-default value")
        self.assertEqual(want,got)
        tree = ET.parse(self.tempdir+'/'+got['xml'][0])
        stylesheets=tree.find("stylesheets").getchildren()
        self.assertEqual(len(stylesheets),1)
        self.assertEqual(stylesheets[0].attrib['href'],'rsrc/css/combined.css')
        javascripts=tree.find("javascripts").getchildren()
        self.assertEqual(len(javascripts),1)
        self.assertEqual(javascripts[0].attrib['src'],'rsrc/js/combined.js')
        self.assertEqual(tree.find("sitename").text,"Blahonga")
        self.assertEqual(tree.find("sitedescription").text,"A non-default value")
        self.assertTrue(os.path.exists(self.tempdir+'/rsrc/css/combined.css'))
        self.assertTrue(os.path.exists(self.tempdir+'/rsrc/js/combined.js'))
        # check that the combining/minifying indeed saved us some space
        # physical path for these: relative to the location of ferenda/manager.py.
        self.assertLess(os.path.getsize(self.tempdir+'/rsrc/css/combined.css'),
                        sum([os.path.getsize(x) for x in ("ferenda/res/css/normalize-1.1.3.css",
                                                          "ferenda/res/css/main.css")]))
        self.assertLess(os.path.getsize(self.tempdir+'/rsrc/js/combined.js'),
                        sum([os.path.getsize(x) for x in ("ferenda/res/js/jquery-1.10.2.js",
                                                          "ferenda/res/js/modernizr-2.6.3.js",
                                                          "ferenda/res/js/respond-1.3.0.js")]))
        # Test3: No combining, make sure that a non-customized
        # DocumentRepository works
        repo = DocumentRepository()
        # but remove any external urls -- that's tested separately in Test5
        repo.config.cssfiles = [x for x in repo.config.cssfiles if not x.startswith("http://")]
        got = manager.makeresources([repo],self.tempdir+os.sep+'rsrc')
        s = os.sep
        want = {'css':[s.join(['rsrc', 'css','normalize-1.1.3.css']),
                       s.join(['rsrc', 'css','main.css']),
                       s.join(['rsrc', 'css','ferenda.css'])],
                'js':[s.join(['rsrc', 'js','jquery-1.10.2.js']),
                      s.join(['rsrc', 'js','modernizr-2.6.3.js']),
                      s.join(['rsrc', 'js','respond-1.3.0.js']),
                      s.join(['rsrc', 'js','ferenda.js'])],
                'xml':[s.join(['rsrc', 'resources.xml'])]
                      }
        self.assertEqual(want,got)

        # test4: Make sure staticsite works (ie no search form in resources.xml):
        repo = DocumentRepository()
        got = manager.makeresources([repo],self.tempdir+os.sep+'rsrc', staticsite = True)
        tree = ET.parse(self.tempdir+os.sep+got['xml'][0])
        search=tree.find("search")
        self.assertFalse(search)

        # test5: include one external resource, combine=False
        test = staticmockclass()
        test.config.cssfiles.append('http://example.org/css/main.css')
        want = {'css':[s.join(['rsrc', 'css','test.css']),
                       'http://example.org/css/main.css'],
                'js':[s.join(['rsrc', 'js','test.js'])],
                'xml':[s.join(['rsrc', 'resources.xml'])]
        }
        got = manager.makeresources([test],self.tempdir+os.sep+'rsrc')
        self.assertEqual(want,got)
                                    
        # test6: include one external resource but with combine=True, which is unsupported
        with self.assertRaises(errors.ConfigurationError):
            got = manager.makeresources([test],self.tempdir+os.sep+'rsrc', combine=True)

        # test7: test the footer() functionality
        test = staticmockclass3()
        got = manager.makeresources([test], self.tempdir+os.sep+'rsrc')
        tree = ET.parse(self.tempdir+os.sep+got['xml'][0])
        footerlinks=tree.findall("footerlinks/nav/ul/li")
        self.assertTrue(footerlinks)
        self.assertEqual(3,len(footerlinks))

        # test8: test win32 path generation on all OS:es, including one full URL
        test = staticmockclass()
        test.config.cssfiles.append('http://example.org/css/main.css')
        want = {'css':['rsrc\\css\\test.css',
                       'http://example.org/css/main.css'],
                'js':['rsrc\\js\\test.js'],
                'xml':['rsrc\\resources.xml']}
        try:
            realsep = os.sep
            os.sep = "\\"
            got = manager.makeresources([test], self.tempdir+os.sep+'rsrc')
            self.assertEqual(want,got)
        finally:
            os.sep = realsep
            
        # test9: nonexistent resources should not be included
        test = staticmockclass()
        test.config.cssfiles = ['nonexistent.css']
        want = {'css':[],
                'js':[s.join(['rsrc', 'js','test.js'])],
                'xml':[s.join(['rsrc', 'resources.xml'])]
        }
        got = manager.makeresources([test], self.tempdir+os.sep+'rsrc')
        self.assertEqual(want,got)
        
        # test10: scss files should be transformed to css
        # disabled until pyScss is usable on py3 again
        # test = staticmockclass()
        # test.config.cssfiles[0] = test.config.cssfiles[0].replace("test.css", "transformed.scss")
        # want = {'css':[s.join(['rsrc', 'css','transformed.css'])],
        #        'js':[s.join(['rsrc', 'js','test.js'])],
        #        'xml':[s.join(['rsrc', 'resources.xml'])]
        # }
        # got = manager.makeresources([test], self.tempdir+os.sep+'rsrc')
        # self.assertEqual(want,got)


    def test_frontpage(self):
        test = staticmockclass()
        test2 = staticmockclass2()
        outfile = self.tempdir+'/index.html'
        manager.makeresources([test,test2], self.tempdir+'/rsrc')
        res = manager.frontpage([test,test2],
                                path=outfile)
        self.assertTrue(res)
        tree = ET.parse(outfile)
        header = tree.find(".//header/h1/a")
        self.assertEqual(header.get("href"), 'http://localhost:8000/')
        # FIXME: check that tree contains 2 divs, that they have id
        # staticmock and staticmock2, that the p text is "Handles
        # foaf:Document documents. Contains 3 published documents."
        divs = tree.findall(".//div[@class='section-wrapper']")
        self.assertEqual(2, len(list(divs)))
        self.assertEqual("staticmock", divs[0].get("id"))
        self.assertEqual("staticmock2", divs[1].get("id"))
        self.assertIn("Handles foaf:Document", divs[0].find("p").text)
        self.assertIn("Contains 3 published documents", divs[0].find("p").text)

    def test_frontpage_staticsite(self):
        test = staticmockclass(datadir=self.tempdir)
        test2 = staticmockclass2(datadir=self.tempdir)
        outfile = self.tempdir+'/index.html'
        manager.makeresources([test,test2], self.tempdir+'/rsrc')
        manager.frontpage([test,test2],
                          path=outfile,
                          staticsite=True)
        # print("\n============== OUTFILE =====================")
        # print(util.readfile(outfile))
        # print("==============================================")
        t = ET.parse(outfile)
        header = t.find(".//header/h1/a")
        self.assertEqual(header.get("href"), 'index.html')

        headernavlinks = t.findall(".//header/nav/ul/li/a")
        self.assertEqual(headernavlinks[0].get("href"), 'staticmock/toc/index.html')
        self.assertEqual(headernavlinks[1].get("href"), 'staticmock2/toc/index.html')

        css = t.findall("head/link[@rel='stylesheet']")
        self.assertRegex(css[0].get('href'), '^rsrc/css')
        
class Setup(RepoTester):

    @patch('ferenda.manager.setup_logger')
    def test_setup(self, mockprint):
        # restart the log system since setup() will do that otherwise
        manager.shutdown_logger()
        manager.setup_logger('CRITICAL')
        projdir = self.datadir+os.sep+'myproject'
        argv= ['ferenda-build.py', projdir]
        
        # test1: normal, setup succeeds
        res = manager.setup(force=True, verbose=False, unattended=True,
                            argv=argv)
        self.assertTrue(res)
        self.assertTrue(os.path.exists(projdir))

        # test2: directory exists, setup fails
        res = manager.setup(verbose=False, unattended=True,
                            argv=argv)
        self.assertFalse(res)
        shutil.rmtree(projdir)
        
        # test2: no argv, rely on sys.argv, assert False
        with patch('ferenda.manager.sys.argv'):
            self.assertFalse(manager.setup())
            self.assertFalse(os.path.exists(projdir))

        # test3: preflight fails
        with patch('ferenda.manager._preflight_check', return_value=False):
            self.assertFalse(manager.setup(unattended=True, argv=argv))
            self.assertFalse(os.path.exists(projdir))

            with patch('ferenda.manager.input', return_value="n") as input_mock:
                self.assertFalse(manager.setup(unattended=False, argv=argv))
                self.assertFalse(os.path.exists(projdir))
                self.assertTrue(input_mock.called)

        # test4: select_triplestore fails
        with patch('ferenda.manager._preflight_check', return_value=True):
            with patch('ferenda.manager._select_triplestore', return_value=(False, None, None)):
                self.assertFalse(manager.setup(unattended=True, argv=argv))
                self.assertFalse(os.path.exists(projdir))

                with patch('ferenda.manager.input', return_value="n") as input_mock:
                    self.assertFalse(manager.setup(unattended=False, argv=argv))
                    self.assertFalse(os.path.exists(projdir))
                    self.assertTrue(input_mock.called)

    def test_preflight(self):
        log = Mock()
        
        # test 1: python too old

        with patch('ferenda.manager.sys') as sysmock:
            sysmock.version_info = (2,5,6,'final',0)
            sysmock.version = sys.version
            self.assertFalse(manager._preflight_check(log, verbose=True))
            self.assertTrue(log.error.called)
            log.error.reset_mock()

        # test 2: modules are old / or missing
        with patch(builtins + '.__import__') as importmock:
            setattr(importmock.return_value, '__version__', '0.0.1')
            self.assertFalse(manager._preflight_check(log, verbose=True))
            self.assertTrue(log.error.called)
            log.error.reset_mock()

            importmock.side_effect = ImportError
            self.assertFalse(manager._preflight_check(log, verbose=True))
            self.assertTrue(log.error.called)
            log.error.reset_mock()

        # test 3: binaries are nonexistent or errors
        with patch('ferenda.manager.subprocess.call') as callmock:
            callmock.return_value = 127
            self.assertFalse(manager._preflight_check(log, verbose=True))
            self.assertTrue(log.error.called)
            log.error.reset_mock()

            callmock.side_effect = OSError
            self.assertFalse(manager._preflight_check(log, verbose=True))
            self.assertTrue(log.error.called)
            log.error.reset_mock()
            
    def test_select_triplestore(self):
        log = Mock()
        # first manipulate requests.get to give the impression that
        # fuseki or sesame either is or isn't available
        with patch('ferenda.manager.requests.get') as mock_get:
            r = manager._select_triplestore("sitename", log, verbose=True)
            self.assertEqual("FUSEKI", r[0])
            
            mock_get.side_effect = requests.exceptions.HTTPError
            r = manager._select_triplestore("sitename", log, verbose=True)
            self.assertNotEqual("FUSEKI", r[0])

            def get_sesame(url):
                if not 'openrdf-sesame' in url:
                    raise requests.exceptions.HTTPError
                resp = Mock()
                resp.text = "ok"
                return resp

            mock_get.side_effect = get_sesame
            r = manager._select_triplestore("sitename", log, verbose=True)
            self.assertEqual("SESAME", r[0])

            mock_get.side_effect = requests.exceptions.HTTPError
            r = manager._select_triplestore("sitename", log, verbose=True)
            self.assertNotEqual("SESAME", r[0])

            # all request.get calls still raises HTTP error
            with patch('ferenda.manager.TripleStore.connect') as mock_connect:
                r = manager._select_triplestore("sitename", log, verbose=True)
                self.assertEqual("SQLITE", r[0])
                def connectfail(storetype, location, repository):
                    if storetype == "SQLITE":
                        raise ImportError("BOOM")
                mock_connect.side_effect = connectfail
                r = manager._select_triplestore("sitename", log, verbose=True)
                self.assertNotEqual("SQLITE", r[0])

                r = manager._select_triplestore("sitename", log, verbose=True)
                self.assertEqual("SLEEPYCAT", r[0])
                mock_connect.side_effect = ImportError
                r = manager._select_triplestore("sitename", log, verbose=True)
                self.assertEqual(None, r[0])
                
    def test_select_fulltextindex(self):
        log = Mock()
        # first manipulate requests.get to give the impression that
        # elasticsearch either is or isn't available
        with patch('ferenda.manager.requests.get') as mock_get:
            r = manager._select_fulltextindex(log, "mysite", verbose=True)
            self.assertEqual("ELASTICSEARCH", r[0])
            self.assertEqual("http://localhost:9200/mysite/", r[1])
            mock_get.side_effect = requests.exceptions.HTTPError

            r = manager._select_fulltextindex(log, "mysite", verbose=True)
            self.assertEqual("WHOOSH", r[0])
            

    def test_runsetup(self):
        with patch('ferenda.manager.sys.exit') as mockexit:
            with patch('ferenda.manager.setup', return_value=True):
                manager.runsetup()
                self.assertFalse(mockexit.called)
                mockexit.reset_mock()
            with patch('ferenda.manager.setup', return_value=False):
                manager.runsetup()
                self.assertTrue(mockexit.called)
                

class Run(unittest.TestCase):
    """Tests manager interface using only the run() entry point used by ferenda-build.py"""

    def setUp(self):
        self.addTypeEqualityFunc(OrderedDict, self.assertDictEqual)
        self.tempdir = tempfile.mkdtemp()
        # self.modulename = hashlib.md5(self.tempdir.encode('ascii')).hexdigest()
        self.modulename = "example"
        self.orig_cwd = os.getcwd()
        # 1. create new blank ini file (FIXME: can't we make sure that
        # _find_config_file is called with create=True when using
        # run() ?)
        os.chdir(self.tempdir)
        util.writefile("ferenda.ini", """[__root__]
loglevel=WARNING
datadir = %s
url = http://localhost:8000
searchendpoint = /search/
apiendpoint = /api/
cssfiles = ['test.css', 'other.css']        
jsfiles = ['test.js']
        """ % self.tempdir)

        # 2. dump 2 example docrepo classes to example.py
        # FIXME: should we add self.tempdir to sys.path also (and remove it in teardown)?
        util.writefile(self.modulename+".py", """# Test code
from ferenda import DocumentRepository, DocumentStore, decorators, errors

class Teststore(DocumentStore):
    def list_basefiles_for(cls,action):
        return ["arg1","myarg","arg2"]
        
class Testrepo(DocumentRepository):
    alias = "test"
    documentstore_class = Teststore
        
    def get_default_options(self):
        opts = super(Testrepo, self).get_default_options()
        opts.update({'datadir': 'data',
                     'cssfiles': ['test.css'],
                     'jsfiles': ['test.js'],
                     'magic': 'less'})
        return opts

    # for inspecting the attributes of a docrepo instance
    @decorators.action
    def inspect(self, attr, subattr=None):
        a = getattr(self,attr)
        if subattr:
            return getattr(a, subattr)
        else:
            return a

    # general testing of arguments and return values (or lack thereof)
    @decorators.action
    def mymethod(self, arg):
        if arg == "myarg":
            return "ok!"

    @decorators.action
    def errmethod(self, arg):
        if arg == "arg1":
            raise Exception("General error")
        elif arg == "myarg":
            raise errors.DocumentRemovedError("Document was removed")
        elif arg == "arg2":
            e = errors.DocumentRemovedError("Document was removed")
            e.dummyfile = "dummyfile.txt"
            raise e

    def download(self):
        return "%s download ok (magic=%s)" % (self.alias, self.config.magic)

    def parse(self, basefile):
        return "%s parse %s" % (self.alias, basefile)

    def relate(self, basefile, otherrepos=[]):
        return "%s relate %s" % (self.alias, basefile)

    def generate(self, basefile, otherrepos=[]): 
        return "%s generate %s" % (self.alias, basefile)

    def toc(self, otherrepos=[]): 
        return "%s toc ok" % (self.alias)

    def news(self, otherrepos=[]): 
        return "%s news ok" % (self.alias)

    @classmethod
    def setup(cls, action, config): pass

    @classmethod
    def teardown(cls, action, config): pass

class CustomStore(DocumentStore):
    def custommethod(self):
        return "CustomStore OK"

    def list_basefiles_for(cls,action):
        return ["arg1","myarg","arg2"]

class Testrepo2(Testrepo):
    alias = "test2"
    storage_policy = "dir"
    downloaded_suffix = ".txt"
    documentstore_class = CustomStore

    @decorators.action
    def mymethod(self, arg):
        if arg == "myarg":
            return "yeah!"

    @decorators.action
    def callstore(self):
        return self.store.custommethod()

""")
        
        util.writefile(self.tempdir+"/test.js", "// test.js code goes here")
        util.writefile(self.tempdir+"/test.css", "/* test.css code goes here */")
        util.writefile(self.tempdir+"/other.css", "/* other.css code goes here */")
        sys.path.append(self.tempdir)

    def tearDown(self):
        manager.shutdown_logger()
        os.chdir(self.orig_cwd)
        shutil.rmtree(self.tempdir)
        sys.path.remove(self.tempdir)


    def test_noconfig(self):
        os.unlink("ferenda.ini")
        with self.assertRaises(errors.ConfigurationError):
            manager.run(["test", "mymethod", "myarg"])
        
    # functionality used by most test methods except test_noconfig
    def _enable_repos(self):

        # 3. run('example.Testrepo', 'enable')
        with patch.object(logging.Logger, 'info') as mocklog:
            self.assertEqual("test",
                             manager.run([self.modulename+".Testrepo", "enable"]))
            # 4. verify that "alias foo enabled" is logged
            log_arg = mocklog.call_args[0][0]
            self.assertEqual("Enabled class %s.Testrepo (alias 'test')" % self.modulename,
                             log_arg)

            # 5. verify that ferenda.ini has changed
            cfg = configparser.ConfigParser()
            cfg.read(["ferenda.ini"])
            self.assertEqual(cfg.get("test","class"), self.modulename+".Testrepo")

            #  (same, with 'example.Testrepo2')
            self.assertEqual("test2",
                             manager.run([self.modulename+".Testrepo2", "enable"]))
            cfg = configparser.ConfigParser()
            cfg.read(["ferenda.ini"])
            self.assertEqual(cfg.get("test2","class"), self.modulename+".Testrepo2")

        with patch.object(logging.Logger, 'error') as mocklog:
            # 6. run('example.Nonexistent', 'enable') -- the ImportError must
            # be caught and an error printed.
            manager.run([self.modulename+".Nonexistent", "enable"])
            # 7. verify that a suitable error messsage is logged
            self.assertEqual("No class named '%s.Nonexistent'" % self.modulename,
                             mocklog.call_args[0][0])

    def test_run_enable(self):
        self._enable_repos()

    def test_run_single(self):
        # test1: run standard (custom) method
        self._enable_repos()
        argv = ["test","mymethod","myarg"]
        self.assertEqual(manager.run(argv),
                         "ok!")
        # test2: specify invalid alias
        argv[0] = "invalid"

        with patch('ferenda.manager.setup_logger'):
            self.assertEqual(manager.run(argv), None)

        with patch(builtins+'.print') as printmock:
            with patch('ferenda.manager.setup_logger'):
                # test3: specify invalid method
                argv = ["test", "invalid"]
                self.assertEqual(manager.run(argv), None)

                # test4: specify no method
                argv = ["test"]
                self.assertEqual(manager.run(argv), None)

    def test_run_single_errors(self):
        self._enable_repos()
        argv = ["test", "errmethod", "--all"]
        with patch('ferenda.manager.setup_logger'):
            with patch(builtins+'.print') as printmock:
                res = manager.run(argv)
        self.assertEqual(res[0][0], Exception)
        self.assertEqual(res[1][0], errors.DocumentRemovedError)
        self.assertEqual(res[2], None)
        self.assertTrue(os.path.exists("dummyfile.txt"))
        
    def test_run_single_all(self):
        self._enable_repos()
        argv = ["test","mymethod","--all"]
        with patch("example.Testrepo.setup", return_value=False):
            self.assertEqual(manager.run(argv), [])

    def test_run_all(self):
        self._enable_repos()
        argv = ["all","mymethod","myarg"]
        self.assertEqual(manager.run(argv),
                         ["ok!", "yeah!"])

        
    def test_run_all_all(self):
        self._enable_repos()
        argv = ["all", "mymethod", "--all"]
        # FIXME: Check to see if proper setup and teardown on Testrepo/Testrepo2 was properly called (how?)
        self.assertEqual(manager.run(argv),
                         [[None,"ok!",None],
                          [None,"yeah!",None]])

    # FIXME: This test magically fails every *other* run on Travis
    # with the following stacktrace:
    # 
    # Traceback (most recent call last):
    #   File "/home/travis/build/staffanm/ferenda/test/testManager.py", line 482, in test_run_all_allmethods
    #     got = manager.run(argv)
    #   File "ferenda/manager.py", line 694, in run
    #     results[action] = run(argscopy)
    #   File "ferenda/manager.py", line 681, in run
    #     return frontpage(**args)
    #   File "ferenda/manager.py", line 360, in frontpage
    #     xsltdir = repos[0].setup_transform_templates(os.path.dirname(stylesheet), stylesheet)
    #   File "ferenda/documentrepository.py", line 1747, in setup_transform_templates
    #     for f in pkg_resources.resource_listdir('ferenda',xsltdir):
    #   File "/home/travis/virtualenv/python2.7/local/lib/python2.7/site-packages/distribute-0.6.34-py2.7.egg/pkg_resources.py", line 932, in resource_listdir
    #     resource_name
    #   File "/home/travis/virtualenv/python2.7/local/lib/python2.7/site-packages/distribute-0.6.34-py2.7.egg/pkg_resources.py", line 1229, in resource_listdir
    #     return self._listdir(self._fn(self.module_path,resource_name))
    #   File "/home/travis/virtualenv/python2.7/local/lib/python2.7/site-packages/distribute-0.6.34-py2.7.egg/pkg_resources.py", line 1320, in _listdir
    #     return os.listdir(path)
    # OSError: [Errno 2] No such file or directory: 'ferenda/res/xsl'
    #
    # I can not figure out why. Skip for now.
    @unittest.skipIf('TRAVIS' in os.environ,
                 "Skipping test_run_all_allmethods on travis-ci")    
    def test_run_all_allmethods(self):
        self._enable_repos()
        argv = ["all", "all", "--magic=more"]
        s = os.sep
        want = OrderedDict(
            [('download', OrderedDict([('test','test download ok (magic=more)'),
                                       ('test2', 'test2 download ok (magic=more)')])),
             ('parse', OrderedDict([('test', ['test parse arg1',
                                              'test parse myarg',
                                              'test parse arg2']),
                                    ('test2', ['test2 parse arg1',
                                               'test2 parse myarg',
                                               'test2 parse arg2'])])),
             ('relate', OrderedDict([('test', ['test relate arg1',
                                               'test relate myarg',
                                               'test relate arg2']),
                                     ('test2', ['test2 relate arg1',
                                                'test2 relate myarg',
                                                'test2 relate arg2'])])),
             ('makeresources', {'css':[s.join(['rsrc', 'css','test.css']),
                                       s.join(['rsrc', 'css','other.css'])],
                                'js':[s.join(['rsrc', 'js','test.js'])],
                                'xml':[s.join(['rsrc', 'resources.xml'])]}),
             ('generate', OrderedDict([('test', ['test generate arg1',
                                                 'test generate myarg',
                                                 'test generate arg2']),
                                    ('test2', ['test2 generate arg1',
                                               'test2 generate myarg',
                                               'test2 generate arg2'])])),
             ('toc', OrderedDict([('test','test toc ok'),
                                ('test2', 'test2 toc ok')])),
             ('news', OrderedDict([('test','test news ok'),
                                   ('test2', 'test2 news ok')])),
             ('frontpage', True)])
        got = manager.run(argv)
        self.maxDiff = None
        self.assertEqual(want,got)
        
    # since this method also calls frontpage, it fails on travis in
    # the same way as test_run_all_allmethods.
    @unittest.skipIf('TRAVIS' in os.environ,
                 "Skipping test_run_single_allmethods on travis-ci")    
    def test_run_single_allmethods(self):
        self._enable_repos()
        argv = ["test","all"]
        s = os.sep
        self.maxDiff = None
        want = OrderedDict(
            [('download', OrderedDict([('test','test download ok (magic=less)'),
                                   ])),
             ('parse', OrderedDict([('test', ['test parse arg1',
                                              'test parse myarg',
                                              'test parse arg2']),
                                ])),
             ('relate', OrderedDict([('test', ['test relate arg1',
                                               'test relate myarg',
                                               'test relate arg2']),
                                 ])),
             ('makeresources', {'css':[s.join(['rsrc', 'css','test.css']),
                                       s.join(['rsrc', 'css','other.css'])],
                                'js':[s.join(['rsrc', 'js','test.js'])],
                                'xml':[s.join(['rsrc', 'resources.xml'])]}),
             ('generate', OrderedDict([('test', ['test generate arg1',
                                                 'test generate myarg',
                                                 'test generate arg2']),
                                   ])),
             ('toc', OrderedDict([('test','test toc ok'),
                              ])),
             ('news', OrderedDict([('test','test news ok'),
                               ])),
            ('frontpage', True)])

        self.assertEqual(manager.run(argv),
                         want)
        
    def test_run_makeresources(self):
        # 1. setup test_run_enable
        # 2. run('all', 'makeresources')
        # 3. verify that all css/jss files specified by default and in Testrepo gets copied
        #    (remove rsrc)
        # 4. run('all', 'makeresources', '--combine')
        # 5. verify that single css and js file is created
        self._enable_repos()
        s = os.sep
        want = {'css':[s.join(['rsrc', 'css','test.css']),
                       s.join(['rsrc', 'css','other.css'])],
                'js':[s.join(['rsrc', 'js','test.js'])],
                'xml':[s.join(['rsrc', 'resources.xml'])]
        }
        got = manager.run(['all', 'makeresources'])
        self.assertEqual(want,got)

        # 6. alter the ferenda.ini so that it doesn't specify any css/js files
        util.writefile("ferenda.ini", """[__root__]
loglevel=WARNING
datadir = %s
url = http://localhost:8000
searchendpoint = /search/
apiendpoint = /api/
        """ % self.tempdir)
        want = {'css':[],
                'js':[],
                'xml':[s.join(['rsrc', 'resources.xml'])]
        }
        got = manager.run(['all', 'makeresources'])
        self.assertEqual(want,got)

    def test_delayed_config(self):
        # Make sure configuration values gets stored properly in the instance created by run()
        self._enable_repos()

        self.assertEqual(".html",               manager.run(['test','inspect','downloaded_suffix']))
        self.assertEqual("file",                manager.run(['test','inspect','storage_policy']))
        self.assertEqual(self.tempdir,          manager.run(['test','inspect','config', 'datadir']))
        self.assertEqual(".html",               manager.run(['test','inspect','store', 'downloaded_suffix']))
        self.assertEqual("file",                manager.run(['test','inspect','store', 'storage_policy']))
        self.assertEqual(self.tempdir+os.sep+"test",  manager.run(['test','inspect','store', 'datadir']))
        
        self.assertEqual(".txt",                 manager.run(['test2','inspect','downloaded_suffix']))
        self.assertEqual("dir",                  manager.run(['test2','inspect','storage_policy']))
        self.assertEqual(self.tempdir,           manager.run(['test2','inspect','config', 'datadir']))
        self.assertEqual(".txt",                 manager.run(['test2','inspect','downloaded_suffix']))
        self.assertEqual("dir",                  manager.run(['test2','inspect','storage_policy']))
        self.assertEqual(self.tempdir+os.sep+"test2",  manager.run(['test2','inspect','store', 'datadir']))
        
    def test_custom_docstore(self):
        self._enable_repos()
        got = manager.run(['test2', 'callstore'])
        self.assertEqual("CustomStore OK", got)

    def test_named_logfile(self):
        self._enable_repos()
        self.assertFalse(os.path.exists("out.log"))
        argv = ["test","mymethod","myarg","--logfile=out.log"]
        manager.run(argv)
        self.assertTrue(os.path.exists("out.log"))
        os.unlink("out.log")

    def test_print_usage(self):
        builtins = "__builtin__" if six.PY2 else "builtins"
        self._enable_repos()
        with patch(builtins+'.print') as printmock:
            manager.run([])

        executable = sys.argv[0]
        got = "\n".join([x[1][0] for x in printmock.mock_calls])
        got = got.replace(executable, "[EXEC]")
        want = """Usage: [EXEC] [class-or-alias] [action] <arguments> <options>
   e.g. '[EXEC] ferenda.sources.EurlexCaselaw enable'
        '[EXEC] ecj parse 62008J0042'
        '[EXEC] all generate'
Available modules:
 * test: [Undocumented]
 * test2: [Undocumented]"""
        self.assertEqual(got, want)
        
    def test_runserver(self):
        self._enable_repos()
        m = Mock()
        with patch('ferenda.manager.make_server', return_value=m) as m2:
            manager.run(["all", "runserver"])
            self.assertTrue(m2.called)
            self.assertTrue(m.serve_forever.called)
        
        

import doctest
from ferenda import manager
def shutup_logger(dt):
    manager.setup_logger('CRITICAL')

def load_tests(loader,tests,ignore):
    tests.addTests(doctest.DocTestSuite(manager, setUp=shutup_logger))
    return tests



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
# Intermediate base class containing some small functionality useful
# for handling data sources of swedish law.

from datetime import datetime, date
import difflib
import os
import re

from rdflib import URIRef, RDF, RDFS, Graph, Namespace
from six import text_type as str

from ferenda import DocumentRepository, DocumentStore, FSMParser, CitationParser
from ferenda import util
from ferenda.sources.legal.se.legalref import Link
from ferenda.elements import Paragraph, Section, Body, CompoundElement, SectionalElement
from ferenda.pdfreader import Page

from . import RPUBL
DCT = Namespace(util.ns['dct'])
PROV = Namespace(util.ns['prov'])

class Stycke(Paragraph):
    pass


class Sektion(Section):
    pass


class PreambleSection(CompoundElement):
    tagname = "div"
    classname = "preamblesection"
    counter = 0
    uri = None
    def as_xhtml(self, uri):
        if not self.uri:
            self.__class__.counter += 1
            self.uri = uri + "#PS%s" % self.__class__.counter
        element = super(PreambleSection, self).as_xhtml(uri)
        element.set('property', 'dct:title')
        element.set('content', self.title)
        element.set('typeof', 'bibo:DocumentPart')
        return element

class UnorderedSection(CompoundElement):
    tagname = "div"
    classname = "unorderedsection"
    counter = 0
    uri = None
    def as_xhtml(self, uri):
        if not self.uri:
            self.__class__.counter += 1
            # note that this becomes a document-global running counter
            self.uri = uri + "#US%s" % self.__class__.counter
        element = super(UnorderedSection, self).as_xhtml(uri)
        element.set('property', 'dct:title')
        element.set('content', self.title)
        element.set('typeof', 'bibo:DocumentPart')
        return element

class Appendix(SectionalElement): 
    tagname = "div"
    classname = "appendix"
    def as_xhtml(self, uri):
        if not self.uri:
            self.uri = uri + "#B%s" % self.ordinal

        return super(Appendix, self).as_xhtml(uri)

class Coverpage(CompoundElement):
    tagname = "div"
    classname = "coverpage"


class SwedishLegalStore(DocumentStore):

    """Customized DocumentStore."""

    def basefile_to_pathfrag(self, basefile):
        # "2012/13:152" => "2012-13/152"
        # "2012:152"    => "2012/152"
        return basefile.replace("/", "-").replace(":", "/")

    def pathfrag_to_basefile(self, pathfrag):
        # "2012-13/152" => "2012/13:152"
        # "2012/152"    => "2012:152"
        return pathfrag.replace("/", ":").replace("-", "/")

    def intermediate_path(self, basefile, attachment=None):
        return self.path(basefile, "intermediate", ".xml", attachment=attachment)


class SwedishLegalSource(DocumentRepository):
    documentstore_class = SwedishLegalStore
    namespaces = ['rdf', 'rdfs', 'xsd', 'dct', 'skos', 'foaf',
                  'xhv', 'xsi', 'owl', 'prov', 'bibo',
                  ('rpubl', 'http://rinfo.lagrummet.se/ns/2008/11/rinfo/publ#'),
                  ('rinfoex', 'http://lagen.nu/terms#')]
    lang="sv"

    rdf_type = RPUBL.Rattsinformationsdokument # subclasses override this

    # This is according to the RPUBL vocabulary: All
    # rpubl:Rattsinformationsdokument should have dct:title,
    # dct:issued (must be a xsd:date), dct:publisher and
    # dct:identifier
    required_predicates = [RDF.type, DCT.title, DCT.issued, DCT.identifier, PROV.wasGeneratedBy]

    swedish_ordinal_list = ('f\xf6rsta', 'andra', 'tredje', 'fj\xe4rde',
                            'femte', 'sj\xe4tte', 'sjunde', '\xe5ttonde',
                            'nionde', 'tionde', 'elfte', 'tolfte')
    swedish_ordinal_dict = dict(list(zip(
        swedish_ordinal_list, list(range(1, len(swedish_ordinal_list) + 1)))))

    swedish_months = {"januari": 1,
                      "februari": 2,
                      "mars": 3,
                      "april": 4,
                      "maj": 5,
                      "juni": 6,
                      "juli": 7,
                      "augusti": 8,
                      "september": 9,
                      "oktober": 10,
                      "november": 11,
                      "december": 12,
                      "\xe5r": 12}

    def get_default_options(self):
        resource_path = os.path.normpath(
            os.path.dirname(__file__) + "../../../../res/etc/authrec.n3")
        opts = super(SwedishLegalSource, self).get_default_options()
        opts['authrec'] = resource_path
        opts['pdfimages'] = False 
        return opts

    def _swedish_ordinal(self, s):
        sl = s.lower()
        if sl in self.swedish_ordinal_dict:
            return self.swedish_ordinal_dict[sl]
        return None

    def _load_resources(self, resource_path):
        # returns a mapping [resource label] => [resource uri]
        # resource_path is given relative to cwd
        graph = Graph()
        graph.load(resource_path, format='n3')
        d = {}
        for uri, label in graph.subject_objects(RDFS.label):
            d[str(label)] = str(uri)
        return d

    def lookup_resource(self, resource_label, cutoff=0.8, warn=True):
        """Given a text label refering to some kind of organization,
        person or other entity, eg. 'Justitiedepartementet Gransk',
        return a URIRef for that entity. The text label does not need to
        match exactly byte-for-byte, a fuzziness matching function
        returns any reasonably similar (adjusted by the cutoff
        parameter) entity."""
        keys = []
        if not hasattr(self, 'org_resources'):
            self.org_resources = self._load_resources(self.config.authrec)

        for (key, value) in list(self.org_resources.items()):
            if resource_label.lower().startswith(key.lower()):
                return URIRef(value)
            else:
                keys.append(key)

        fuzz = difflib.get_close_matches(resource_label, keys, 1, cutoff)
        if fuzz:
            if warn:
                self.log.warning("Assuming that '%s' should be '%s'?" %
                                 (resource_label, fuzz[0]))
            return URIRef(self.lookup_resource(fuzz[0]))
        else:
            self.log.warning("No good match for '%s'" % (resource_label))
            raise KeyError(resource_label)

    def lookup_label(self, resource):
        if not hasattr(self, 'org_resources'):
            self.org_resources = self._load_resources(self.config.authrec)
        for (key, value) in list(self.org_resources.items()):
            if resource == value:
                return key

        raise KeyError(resource)

    def sameas_uri(self, uri):
        # "http://localhost:8000/res/dir/2012:35" => "http://rinfo.lagrummet.se/publ/dir/2012:35",
        # "http://localhost:8000/res/dv/hfd/2012:35" => "http://rinfo.lagrummet.se/publ/rattsfall/hdf/2012:35",
        assert uri.startswith(self.config.url)
        # FIXME: This hardcodes the res/ part of our local URIs
        # needlessly -- make configurable
        maps = (("res/dv/", "publ/rattsfall/"),
                ("res/", "publ/"))
        for fr, to in maps:
            if self.config.url + fr in uri:
                return uri.replace(self.config.url + fr,
                                   "http://rinfo.lagrummet.se/" + to)

    def parse_iso_date(self, datestr):
        # only handles YYYY-MM-DD now. Look into dateutil or isodate
        # for more complete support of all ISO 8601 variants
        return datetime.strptime(datestr, "%Y-%m-%d")

    def parse_swedish_date(self, datestr):
        """ Parses a number of common forms of expressing swedish dates with varying precision.
        >>> parse_swedish_date("3 februari 2010")
        datetime.date(2010, 2, 3)
        >>> parse_swedish_date("vid utgången av december 1999")
        datetime.date(1999, 12, 31)
        >>> parse_swedish_date("november 1999")
        ferenda.util.gYearMonth(1999, 11)
        >>> parse_swedish_date("1998")
        ferenda.util.gYear(1999)
        """
        day = month = year = None
        # assume strings on the form "3 februari 2010"
        # strings on the form "vid utg\xe5ngen av december 1999"
        if datestr.startswith("vid utg\xe5ngen av"):
            import calendar
            (x, y, z, month, year) = datestr.split()
            month = self.swedish_months[month]
            year = int(year)
            day = calendar.monthrange(year, month)[1]
        else:
            # assume strings on the form "3 februari 2010"
            components =  datestr.split()
            year = int(components[-1])
            if len(components) >= 2:
                month = self.swedish_months[components[-2]]
            if len(components) >= 3:
                day = int(components[-3])

        # return the best we can
        if day:
            return date(year, month, day)
        if month:
            return util.gYearMonth(year, month)
        else:
            return util.gYear(year)


    def infer_triples(self, d, basefile):
        try:
            identifier = d.getvalue(self.ns['dct'].identifier)
            # if the identifier is incomplete, eg "2010/11:68" instead
            # of "Prop. 2010/11:68", the following triggers a
            # ValueError, which is handled the same as if no
            # identifier is available at all. Ideally,
            # sanitize_identifier should prevent all preventable
            # occurrences of this.
            (doctype, arsutgava, lopnummer) = re.split("[ :]", identifier)
        except (KeyError, ValueError):
            # Create one from basefile. First guess prefix
            if self.rdf_type == self.ns['rpubl'].Direktiv:
                prefix = "Dir. "
            elif self.rdf_type == self.ns['rpubl'].Utredningsbetankande:
                if d.getvalue(self.ns['rpubl'].utrSerie) == "http://rinfo.lagrummet.se/serie/utr/ds":
                    prefix = "Ds "
                else:
                    prefix = "SOU "
            elif self.rdf_type == self.ns['rpubl'].Proposition:
                prefix = "Prop. "
            elif self.rdf_type == self.ns['rpubl'].Forordningsmotiv:
                prefix = "Fm "
            else:
                raise ValueError("Cannot create dct:identifier for rdf_type %r" % self.rdf_type)
            identifier = "%s%s" % (prefix, basefile)
            
            self.log.warning(
                "%s: No dct:identifier, assuming %s" % (basefile, identifier))
            d.value(self.ns['dct'].identifier, identifier)

        # self.log.debug("Identifier %s" % identifier)
        (doctype, arsutgava, lopnummer) = re.split("[ :]", identifier)
        d.value(self.ns['rpubl'].arsutgava, arsutgava)
        d.value(self.ns['rpubl'].lopnummer, lopnummer)

    def toc_query(self):
        return """PREFIX dct:<http://purl.org/dc/terms/>
                  PREFIX rpubl:<http://rinfo.lagrummet.se/ns/2008/11/rinfo/publ#>
                  SELECT DISTINCT ?uri ?title ?identifier ?arsutgava ?lopnummer ?departement
                  FROM <%s>
                  WHERE {?uri dct:title ?title;
                              dct:identifier ?identifier;
                              rpubl:arsutgava ?arsutgava;
                              rpubl:lopnummer ?lopnummer;
                              rpubl:departement ?departement;
                  }""" % self.context()

    def toc_criteria(self):
        return (
            {'predicate': self.ns['rpubl']['arsutgava'],
             'binding': 'arsutgava',
             'label': 'Efter \xe5rtal',
             'sorter': cmp,
             'pages': []},
            {'predicate': self.ns['dct']['title'],
             'binding': 'title',
             'label': 'Efter rubrik',
             'selector': lambda x: x[0].lower(),
             'sorter': cmp,
             'pages': []},
            {'predicate': self.ns['rpubl']['departement'],
             'binding': 'departement',
             'label': 'Efter departement',
             'selector': self.lookup_label,
             'sorter': cmp,
             'pages': []},
        )

    def toc_item(self, binding, row):
        return {'uri': row['uri'],
                'label': row['identifier'] + ": " + row['title']}

def offtryck_parser(basefile="0", preset="proposition", metrics={}):
    presets = {'default': {},
               'dir': {'footer': 920,
                       'header': 82,
                       'leftmargin': 105,
                       'rightmargin': 566,
                       'headingsize': 14,
                       'subheadingsize': 14,
                       'subheadingfamily': 'TimesNewRomanPS-ItalicMT',
                       'subsubheadingsize': None,
                       'textsize': 14},
               'proposition': {'footer': 920,
                               'header': 65, # make sure this is correct
                               'leftmargin': 160,
                               'rightmargin': 628,
                               'headingsize': 20,
                               'subheadingsize': 17,
                               'subheadingfamily': 'Times New Roman',
                               'subsubheadingsize': 15,
                               'textsize': 13},
               'sou': {'header': 49, # or rather 49 + 15
                       'header': 65, # make sure this is correct
                       'footer': 940,
                       'leftmargin': 84,
                       'rightmargin': 813,
                       'titlesize': 41,
                       'headingsize': 26,
                       'subheadingsize': 16,
                       'subheadingfamily': 'TradeGothic,Bold',
                       'subsubheadingsize': 14,
                       'textsize': 14
                   },
               'ds': {'header': 49, # or rather 49 + 15
                      'header': 65, # make sure this is correct
                      'footer': 940,
                      'leftmargin': 84,
                      'rightmargin': 813,
                      'titlesize': 41,
                      'headingsize': 26,
                      'subheadingsize': 16,
                      'subheadingfamily': 'TradeGothic,Bold',
                      'subsubheadingsize': 14,
                      'textsize': 14
                   }
               }
    if preset:
        metrics = presets[preset]

    # a mutable variable, which is accessible from the nested
    # functions
    state = {'pageno': 0,
             'appendixno': None,
             'preset': preset}

    def is_pagebreak(parser):
        return isinstance(parser.reader.peek(), Page)

    # page numbers, headings.
    def is_nonessential(parser):
        chunk = parser.reader.peek()
        if chunk.top > metrics['footer'] or chunk.bottom < metrics['header']:
            return True  # page numbers
        if (int(chunk.getfont()['size']) <= metrics['textsize'] and
                (chunk.left < metrics['leftmargin'] or
                 chunk.left > metrics['rightmargin']) and
            (15 <= len(str(chunk)) <= 29)): # matches both "Prop. 2013/14:1" and "Prop. 1999/2000:123 Bilaga 12"
            return True

    def is_coverpage(parser):
        # first 2 pages of a SOU are coverpages
        return isinstance(parser.reader.peek(), Page) and state['preset'] == "sou" and state['pageno'] < 2

            
    def is_preamblesection(parser):
        chunk = parser.reader.peek()
        if isinstance(chunk, Page):
            return False
        txt = str(chunk).strip()
        fontsize = int(chunk.getfont()['size'])
        if not metrics['subheadingsize'] <= fontsize <= metrics['headingsize']:
            return False

        for validheading in ('Propositionens huvudsakliga innehåll',
                             'Innehållsförteckning',
                             'Till statsrådet',
                             'Innehåll',
                             'Sammanfattning'):
            if txt.startswith(validheading):
                return True

    def is_section(parser):
        (ordinal, title) = analyze_sectionstart(parser)
        if ordinal:
            return ordinal.count(".") == 0

    def is_subsection(parser):
        (ordinal, title) = analyze_sectionstart(parser)
        if ordinal:
            return ordinal.count(".") == 1

    def is_unorderedsection(parser):
        # Subsections in "Författningskommentar" sections are
        # not always numbered. As a backup, check font size and family as well
        chunk = parser.reader.peek()
        return (int(chunk.getfont()['size']) == metrics['subheadingsize'] and
                chunk.getfont()['family'] == metrics['subheadingfamily'])

    def is_subsubsection(parser):
        (ordinal, title) = analyze_sectionstart(parser)
        if ordinal:
            return ordinal.count(".") == 2

    def is_appendix(parser):
        chunk = parser.reader.peek()
        txt = str(chunk).strip()
        if (chunk.getfont()['size'] == metrics['headingsize'] and txt.startswith("Bilaga ")):
            return True
        elif (int(chunk.getfont()['size']) == metrics['textsize'] and
              (chunk.left < metrics['leftmargin'] or
               chunk.left > metrics['rightmargin'])):
            m = re.search("Bilaga (\d)", str(chunk))
            if m:
                ordinal = int(m.group(1))
                if ordinal != state['appendixno']:
                    return True

    def is_paragraph(parser):
        return True

    def make_body(parser):
        return p.make_children(Body())
    setattr(make_body, 'newstate', 'body')

    def make_paragraph(parser):
        # if "Regeringen beslutade den 8 april 2010 att" in str(parser.reader.peek()):
        #     raise ValueError("OK DONE")
        return parser.reader.next()

    def make_coverpage(parser):
        state['pageno'] += 1
        parser.reader.next() # throwaway the Page object itself
        c = Coverpage()
        return parser.make_children(c)
    setattr(make_coverpage, 'newstate', 'coverpage')
        

    def make_preamblesection(parser):
        s = PreambleSection(title=str(parser.reader.next()).strip())
        if s.title == "Innehållsförteckning":
            parser.make_children(s) # throw away
            return None
        else:
            return parser.make_children(s)
    setattr(make_preamblesection, 'newstate', 'preamblesection')


    def make_unorderedsection(parser):
        s = UnorderedSection(title=str(parser.reader.next()).strip())
        return parser.make_children(s)
    setattr(make_unorderedsection, 'newstate', 'unorderedsection')

    def make_appendix(parser):
        # now, an appendix can begin with either the actual
        # headline-like title, or by the sidenote in the
        # margin. Find out which it is, and plan accordingly.
        done = False
        while not done:
            chunk = parser.reader.next()
            if isinstance(chunk, Page):
                continue
            m = re.search("Bilaga (\d)", str(chunk))
            if m:
                state['appendixno'] = int(m.group(1))
            if int(chunk.getfont()['size']) >= metrics['subheadingsize']:
                done = True
        s = Appendix(title=str(chunk).strip(),
                     ordinal=str(state['appendixno']),
                     uri=None)
        return parser.make_children(s)
    setattr(make_appendix, 'newstate', 'appendix')

    # this is used for subsections and subsubsections as well --
    # probably wont work due to the newstate property
    def make_section(parser):
        ordinal, title = analyze_sectionstart(parser, parser.reader.next())
        if ordinal:
            identifier = "Prop. %s, avsnitt %s" % (basefile, ordinal)
            s = Section(ordinal=ordinal, title=title)
        else:
            s = Section(title=str(title))
        return parser.make_children(s)
    setattr(make_section, 'newstate', 'section')

    def skip_nonessential(parser):
        parser.reader.next()
        return None

    def skip_pagebreak(parser):
        # increment pageno
        state['pageno'] += 1
        parser.reader.next()
        return None

    re_sectionstart = re.compile("^(\d[\.\d]*) +(.*[^\.])$").match
    def analyze_sectionstart(parser, textbox=None):
        if not textbox:
            textbox = parser.reader.peek()
        if not (metrics['headingsize'] >= int(textbox.getfont()['size']) >= metrics['subsubheadingsize']):
            return (None, textbox)
        txt = str(textbox)
        m = re_sectionstart(txt)
        if m:
            ordinal = m.group(1).rstrip(".")
            title = m.group(2)
            return (ordinal, title.strip())
        else:
            return (None, textbox)

    p = FSMParser()

    p.set_recognizers(is_coverpage,
                      is_pagebreak,
                      is_appendix,
                      is_nonessential,
                      is_section,
                      is_subsection,
                      is_subsubsection,
                      is_preamblesection,
                      is_unorderedsection,
                      is_paragraph)
    commonstates = ("body","preamblesection","section", "subsection", "unorderedsection", "subsubsection", "appendix")
    p.set_transitions({(commonstates, is_nonessential): (skip_nonessential, None),
                       (commonstates, is_pagebreak): (skip_pagebreak, None),
                       (commonstates, is_unorderedsection): (make_unorderedsection, "unorderedsection"),
                       (commonstates, is_paragraph): (make_paragraph, None),
                       ("body", is_coverpage): (make_coverpage, "coverpage"),
                       ("body", is_preamblesection): (make_preamblesection, "preamblesection"),
                       ("coverpage", is_coverpage): (False, None),
                       ("coverpage", is_preamblesection): (False, None),
                       ("coverpage", is_paragraph): (make_paragraph, None),
                       ("preamblesection", is_preamblesection): (False, None),
                       ("preamblesection", is_section): (False, None),
                       ("body", is_section): (make_section, "section"),
                       ("section", is_section): (False, None),
                       ("section", is_subsection): (make_section, "subsection"),
                       ("unorderedsection", is_preamblesection): (False, None),
                       ("unorderedsection", is_unorderedsection): (False, None),
                       ("unorderedsection", is_section): (False, None),
                       ("unorderedsection", is_appendix): (False, None),
                       ("subsection", is_subsection): (False, None),
                       ("subsection", is_section): (False, None),
                       ("subsection", is_subsubsection): (make_section, "subsubsection"),
                       ("subsubsection", is_subsubsection): (False, None),
                       ("subsubsection", is_subsection): (False, None),
                       ("subsubsection", is_section): (False, None),
                       ("body", is_appendix): (make_appendix, "appendix"),
                       (("appendix","subsubsection", "subsection", "section"), is_appendix):
                       (False, None)
                       })

    p.initial_state = "body"
    p.initial_constructor = make_body
    return p


def offtryck_gluefunc(textbox, nextbox, prevbox):
    linespacing = int(nextbox.getfont()['size']) / 2
    parindent = int(nextbox.getfont()['size'])
    if (textbox.getfont()['size'] == nextbox.getfont()['size'] and
        textbox.getfont()['family'] == nextbox.getfont()['family'] and
        textbox.top + textbox.height + linespacing > nextbox.top and
        ((prevbox.top + prevbox.height == nextbox.top + nextbox.height) or # compare baseline, not topline
         (prevbox.left == nextbox.left) or
         (parindent * 2 >= (prevbox.left - nextbox.left) >= parindent)
     )):
     return True
    
# (ab)use the CitationClass, with it's useful parse_recursive method,
# to use a legalref based parser instead of a set of pyparsing
# grammars.
class SwedishCitationParser(CitationParser):
    def __init__(self, legalrefparser):
        self._legalrefparser = legalrefparser

    def parse_string(self, string):
        unfiltered = self._legalrefparser.parse(string, predicate="dct:references")
        # remove those references that we cannot fully resolve (should
        # be an option in LegalRef, but...
        filtered = []
        for node in unfiltered:
            if isinstance(node, Link) and "sfs/9999:999" in node.uri:
                filtered.append(str(node))
            else:
                filtered.append(node)
        return filtered



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from six import text_type as str

from ferenda.elements import LinkSubject

class CitationParser(object):

    """Finds citations to documents and other resources in text
    strings. Each type of citation is specified by a
    `pyparsing <http://pyparsing.wikispaces.com/Documentation>`_
    grammar, and for each found citation a URI can be constructed
    using a :py:class:`~ferenda.URIFormatter` object.
    
    :param grammars: The grammar(s) for the citations that this
                      parser should find, in order of priority.
    :type  grammars: list of ``pyparsing.ParserElement`` objects

    Usage:

    >>> from pyparsing import Word,nums
    >>> rfc_grammar = ("RFC " + Word(nums).setResultsName("rfcnumber")).setResultsName("rfccite")
    >>> pep_grammar = ("PEP" +  Word(nums).setResultsName("pepnumber")).setResultsName("pepcite")
    >>> citparser = CitationParser(rfc_grammar, pep_grammar)
    >>> res = citparser.parse_string("The WSGI spec (PEP 333) references RFC 2616 (The HTTP spec)")
    >>> # res is a list of strings and/or pyparsing.ParseResult objects
    >>> from ferenda import URIFormatter
    >>> from ferenda.elements import Link
    >>> f = URIFormatter(('rfccite',
    ...                   lambda p: "http://www.rfc-editor.org/rfc/rfc%(rfcnumber)s" % p),
    ...                  ('pepcite',
    ...                   lambda p: "http://www.python.org/dev/peps/pep-0%(pepnumber)s/" % p))
    >>> citparser.set_formatter(f)
    >>> res = citparser.parse_recursive(["The WSGI spec (PEP 333) references RFC 2616 (The HTTP spec)"])
    >>> res == ['The WSGI spec (', Link('PEP 333',uri='http://www.python.org/dev/peps/pep-0333/'), ') references ', Link('RFC 2616',uri='http://www.rfc-editor.org/rfc/rfc2616'), ' (The HTTP spec)']
    True
    """

    def __init__(self, *grammars):
        self._grammars = []
        for grammar in grammars:
            self.add_grammar(grammar)
        self._formatter = None

    def set_formatter(self, formatter):
        """Specify how found citations are to be formatted when using
        :py:meth:`~ferenda.CitationParser.parse_recursive`
        
        :param formatter: The formatter object to use for all citations
        :type  formatter: :py:class:`~ferenda.URIFormatter`
        """
        self._formatter = formatter

    def add_grammar(self, grammar):
        """Add another grammar.

        :param grammar: The grammar to add
        :type grammar: ``pyparsing.ParserElement``
        """
        self._grammars.append(grammar)

    def parse_string(self, string):
        """Find any citations in a text string, using the configured grammars.
        
        :param string: Text to parse for citations
        :type string: str
        :returns: strings (for parts of the input text that do not contain
                  any citation) and/or tuples (for found citation) consisting
                  of (string, ``pyparsing.ParseResult``)
        :rtype: list
        """
        # Returns a list of strings and/or tuples, where each tuple is
        # (string,pyparsing.ParseResult)
        nodes = [string]
        for grammar in self._grammars:
            res = []
            for node in nodes:
                if not isinstance(node, str):
                    res.append(node)
                    continue
                matches = grammar.scanString(node)
                start = 0
                after = 0
                for match, before, after in matches:
                    if before > start:
                        res.append(node[start:before])
                    res.append((node[before:after], match))
                    start = after
                if after < len(node):
                    res.append(node[after:])
            nodes = list(res)
        return res

    def parse_recursive(self, part):
        """Traverse a nested tree of elements, finding citations in
        any strings contained in the tree. Found citations are marked
        up as :py:class:`~ferenda.elements.Link` elements with the uri
        constructed by the :py:class:`~ferenda.URIFormatter` set by
        :py:meth:`~ferenda.CitationParser.set_formatter`.
        
        :param part: The root element of the structure to parse
        :type  part: list
        :returns: a correspondingly nested structure.
        :rtype: list"""

        res = []
        if not isinstance(part, str):
            for subpart in part:
                if isinstance(subpart, str):
                    res.extend(self.parse_recursive(subpart))
                else:
                    res.append(self.parse_recursive(subpart))
            # replace our exising subparts/children with the combined result of parse_recursive
            part[:] = res[:]
            return part

        else:  # ok, simple string
            # FIXME: We need to keep track of the URI for the part
            # of the document we're in, so that we can resolve
            # partial/relative references
            # splits a string into a list of string and ParseResult objects
            nodes = self.parse_string(part)
            for node in nodes:
                if isinstance(node, str):
                    res.append(node)
                elif isinstance(node, tuple):
                    (text, parseresult) = node
                    # node = self.resolve_relative(node,currentloc)
                    uri = self._formatter.format(parseresult)
                    if uri:
                        res.append(LinkSubject(
                            text, uri=uri, predicate="dct:references"))
                    else:
                        res.append(text)
                # FIXME: concatenate adjacent str nodes
            return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys
import os
sys.path.append(os.path.normpath(os.getcwd() + os.sep + os.pardir))

from ferenda import manager
manager.run(sys.argv[1:])





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import tempfile
import shutil
import os
import sys

from ferenda import errors
from ferenda.compat import unittest, patch

# SUT
from ferenda import util

class Main(unittest.TestCase):

    def p(self, path, prepend_datadir=True):
        if prepend_datadir:
            path = self.datadir + "/" + path
        return path.replace('/', '\\') if os.sep == '\\' else path

    def setUp(self):
        self.datadir = tempfile.mkdtemp()
        self.dname = self.datadir + os.sep + "foo"
        self.fname = self.datadir + os.sep + "foo/bar.txt"
        self.fname2 = self.datadir + os.sep + "foo/baz.txt"

    def tearDown(self):
        shutil.rmtree(self.datadir)

    def test_ensure_dir(self):
        self.assertFalse(os.path.exists(self.dname))
        util.ensure_dir(self.fname)
        self.assertTrue(os.path.exists(self.dname))
        self.assertTrue(os.path.isdir(self.dname))
        util.ensure_dir(self.fname)
        os.rmdir(self.dname)
        with patch('ferenda.util.mkdir', side_effect=OSError):
            util.ensure_dir(self.fname)

    def test_robust_rename(self):
        # only test the IOError branch
        util.writefile(self.fname, "Hello")
        util.writefile(self.fname2, "Hello")
        with patch('ferenda.util.shutil.move', side_effect=IOError):
            util.robust_rename(self.fname, self.fname2)

    def test_robust_remove(self):
        util.writefile(self.fname, "Hello")
        util.robust_remove(self.fname)
        util.robust_remove(self.fname)
        
    def test_runcmd(self):
        filename = self.dname+os.sep+"räksmörgås.txt"
        util.writefile(filename, "räksmörgås")
        if sys.platform == "win32":
            cmd = "type"
        else:
            cmd = "cat"
        cmdline = "%s %s" % (cmd, filename)
        (retcode, stdout, stderr) = util.runcmd(cmdline)
        self.assertEqual(0, retcode)
        self.assertEqual("räksmörgås", stdout)
        self.assertEqual("", stderr)
        
        cmdline = "non-existing-binary foo"
        (retcode, stdout, stderr) = util.runcmd(cmdline)
        self.assertNotEqual(0, retcode)
        self.assertNotEqual("", stderr)

        with self.assertRaises(errors.ExternalCommandError):
            (retcode, stdout, stderr) = util.runcmd(cmdline,
                                                    require_success=True)

    def test_listdirs(self):
        util.writefile(self.p("foo.txt"), "Hello")
        util.writefile(self.p("bar.txt"), "Hello")
        util.writefile(self.p("foo/2.txt"), "Hello")
        util.writefile(self.p("foo/10.txt"), "Hello")
        util.writefile(self.datadir+"/foo/baz.text", "Hello")
        generator = util.list_dirs(self.datadir, ".txt")
        self.assertEqual(self.p("bar.txt"), next(generator))
        self.assertEqual([self.p("foo.txt"),
                          self.p("foo/2.txt"),
                          self.p("foo/10.txt")], list(generator))

    def test_replace_if_different(self):
        # test 1: dst does not exist
        util.writefile(self.fname, "Hello")
        self.assertTrue(util.replace_if_different(self.fname, self.fname2))
        self.assertFalse(os.path.exists(self.fname))
        self.assertTrue(os.path.exists(self.fname2))

        # test 2: dst exists, but is different (gets overwritten)
        util.writefile(self.fname, "Hello (different)")
        self.assertTrue(util.replace_if_different(self.fname, self.fname2))
        self.assertFalse(os.path.exists(self.fname))
        self.assertEqual("Hello (different)",
                         util.readfile(self.fname2))

        # test 3: src and dst is identical (src gets removed)
        util.writefile(self.fname, "Hello (different)")
        self.assertFalse(util.replace_if_different(self.fname, self.fname2))
        self.assertFalse(os.path.exists(self.fname))

        # test 4: dst exist, is different, gets archived
        newfile = self.dname+"/new.txt"
        archivefile = self.dname+"/archive.txt"
        util.writefile(newfile, "Hello (archiving)")
        self.assertTrue(util.replace_if_different(newfile, self.fname2, archivefile))
        self.assertFalse(os.path.exists(newfile))
        self.assertEqual("Hello (archiving)",
                         util.readfile(self.fname2))
        self.assertEqual("Hello (different)",
                         util.readfile(archivefile))

    def test_copy_if_different(self):
        # test 1: dst does not exist
        util.writefile(self.fname, "Hello")
        self.assertTrue(util.copy_if_different(self.fname, self.fname2))
        self.assertTrue(os.path.exists(self.fname))
        self.assertTrue(os.path.exists(self.fname2))

        # test 2: dst does exist, is different
        util.writefile(self.fname, "Hello (different)")
        self.assertTrue(util.copy_if_different(self.fname, self.fname2))
        self.assertTrue(os.path.exists(self.fname))
        self.assertTrue(os.path.exists(self.fname2))
        self.assertEqual("Hello (different)",
                         util.readfile(self.fname2))

        # test 3: dst does exist, is identical
        self.assertFalse(util.copy_if_different(self.fname, self.fname2))


from ferenda import util
import doctest
def load_tests(loader,tests,ignore):
    tests.addTests(doctest.DocTestSuite(util))
    return tests




# flake8: noqa
from .citationparser import CitationParser
from .uriformatter import URIFormatter
from .describer import Describer
from .layeredconfig import LayeredConfig
from .pdfreader import PDFReader
from .textreader import TextReader
from .wordreader import WordReader
from .triplestore import TripleStore
from .fulltextindex import FulltextIndex
from .devel import Devel
from .fsmparser import FSMParser
from .tocpageset import TocPageset
from .tocpage import TocPage
from .toccriteria import TocCriteria
from .newscriteria import NewsCriteria
from .transformer import Transformer
from .document import Document
from .documententry import DocumentEntry
from .documentstore import DocumentStore
from .documentrepository import DocumentRepository
from .pdfdocumentrepository import PDFDocumentRepository
from .compositerepository import CompositeRepository, CompositeStore
__version__ = "0.1.7" #gets pulled into setup.py and docs/conf.py



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from operator import attrgetter


class NewsCriteria(object):

    """Represents a particular subset of the documents in a repository,
    for the purpose of generating a news feed for that subset. These
    criteria objects are used by
    :py:meth:`~ferenda.DocumentRepository.news`, and the simplest way
    of controlling which criteria are used for your docrepo is to
    override :py:meth:`~ferenda.DocumentRepository.news_criteria` to
    make it return a list of instantiated NewsCriteria objects.

    :param basefile: A slug-like basic text label for this subset.
    :type  basefile: str
    :param feedtitle: The title for this particular news feed
    :type  feedtitle: str
    :param selector: Function that takes a single
                     :py:class:`~ferenda.DocumentEntry` object and
                     returns true iff it should be included in this
                     feed.
    :type  selector: callable
    :param key: Function that takes a single
                :py:class:`~ferenda.DocumentEntry` object and returns
                a value that can be used for sorting that object. The
                default implementation returns the entrys
                ``updated`` attribute, so that the feed contains entries
                sorted most recently updated first.
    :type  key: callable

    """

    def __init__(self, basefile, feedtitle, selector=None, key=None):
        self.basefile = basefile
        self.feedtitle = feedtitle
        if not selector:
            self.selector = lambda entry: True
        else:
            assert callable(selector)
            self.selector = selector
        if not key:
            self.key = attrgetter('updated')  # or lambda x: x.updated
        else:
            assert callable(key)
            self.key = key
        self.entries = []



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys
import os
import subprocess
import tempfile
import shutil

import six

from ferenda import util
from ferenda.compat import unittest, patch
from ferenda.testutil import FerendaTestCase
# This testcase tests those examples in the documentation that are
# more unit-like and can run without downloading stuff from the
# net. More integration-like tests are in integrationTestExamples (and
# thus aren't run in the normal test suite).

# imports needed by the scripts. I do not fully understand exactly how
# imports are scoped when using exec, but this is the only way apart
# from importing inside of the functions that use the code to work.
from ferenda import elements, DocumentRepository, DocumentStore, TocCriteria
from ferenda.decorators import managedparsing
import ferenda.citationpatterns
import ferenda.uriformats
from bs4 import BeautifulSoup
import requests
from six.moves.urllib_parse import urljoin
XMLPatents = HTMLPatents = ScannedPatents = None

class TestExamples(unittest.TestCase, FerendaTestCase):
    def _test_pyfile(self, pyfile, want=True, comparator=None):
        with open(pyfile, 'rb') as fp:
            pycode = compile(fp.read(), pyfile, 'exec')
        result = six.exec_(pycode, globals(), locals())
        # the exec:ed code is expected to set return_value
        got = locals()['return_value']
        if not comparator:
            comparator = self.assertEqual
        comparator(want, got)

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.orig_cwd = os.getcwd()
        os.chdir(self.tempdir)
        
    def tearDown(self):
        os.chdir(self.orig_cwd)
        shutil.rmtree(self.tempdir)

    def test_elementclasses(self):
        # setup w3standards.py -- modify sys.path?
        self._test_pyfile(self.orig_cwd + "/doc/examples/elementclasses.py",
                          util.readfile(self.orig_cwd + "/doc/examples/elementclasses-part.xhtml", "rb"),
                          self.assertEqualXML)

    def test_fsmparser_example(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/fsmparser-example.py",
                          util.readfile(self.orig_cwd + "/doc/examples/fsmparser-result.xml"),
                          self.assertEqualXML)

    def test_keyconcepts_attachments(self):
        with patch('requests.get'):
            self._test_pyfile(self.orig_cwd + "/doc/examples/keyconcepts-attachments.py")

    def test_keyconcepts_file(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/keyconcepts-file.py")

    def test_metadata(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/metadata.py",
                          util.readfile(self.orig_cwd + "/doc/examples/metadata-result.xml"),
                          self.assertEqualXML)

    def test_citationparsing_urls(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/citationparsing-urls.py")
        
    def test_citationparsing_parsers(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/citationparsing-parsers.py",
                          util.readfile(self.orig_cwd + "/doc/examples/citationparsing-after.xhtml"),
                          self.assertEqualXML)
        
    def test_citationparsing_custom(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/citationparsing-custom.py")

    def test_composite(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/patents.py")

    def test_toc(self):
        self._test_pyfile(self.orig_cwd + "/doc/examples/toc.py")



# -*- coding: utf-8 -*-
"""The purpose of this module is to provide classes corresponding to
most elements (except ``<style>``, ``<script>`` and similar
non-document content elements) and core attributes (except ``@style``
and the ``%events`` attributes) of HTML4.01 and HTML5. It is not
totally compliant with the HTML4.01 and HTML5 standards, but is enough
to model most real-world HTML. It contains no provisions to ensure
that elements of a particular kind only contain allowed
sub-elements.

"""
from __future__ import unicode_literals

import logging
import six

import bs4

from . import CompoundElement


def elements_from_soup(soup,
                       remove_tags=('script', 'style', 'font', 'map', 'center'),
                       keep_attributes=('class', 'id', 'dir', 'lang', 'src', 'href', 'name', 'alt')):
    """Converts a BeautifulSoup tree into a tree of
    :py:class:`ferenda.elements.html.HTMLElement` objects. Some
    non-semantic attributes and tags are removed in the process.

    :param soup: Soup object to convert
    :param remove_tags: Tags that should not be included
    :type  remove_tags: list
    :param keep_attributes: Attributes to keep
    :type  keep_attributes: list
    :returns: tree of element objects
    :rtype: ferenda.elements.html.HTMLElement

    """
    log = logging.getLogger(__name__)
    if soup.name in remove_tags:
        return None
    if soup.name not in _tagmap:
        # self.log.warning("Can't render %s" % soup.name)
        log.warning("Can't render %s" % soup.name)
        return None
    attrs = {}
    for attr in keep_attributes:
        if attr in soup.attrs:
            # print("   %s has attr %s" % (soup.name,attr))
            if isinstance(soup[attr], list):
                attrs[attr] = " ".join(soup[attr])
            else:
                attrs[attr] = soup[attr]
    # print("%s: %r" % (soup.name, attrs))

    element = _tagmap[soup.name](**attrs)

    #print("%sNode: %s" % ((depth-1)*". ",soup.name))
    for child in soup.children:
        if isinstance(child, bs4.element.Comment):
            #print("%sChild comment" % (depth*". "))
            pass
        elif isinstance(child, bs4.NavigableString):
            #print("%sChild string %r" % (depth*". ",child[:10]))
            if six.text_type(child).strip() != "":  # ignore pure whitespace between tags
                element.append(six.text_type(child))  # convert NavigableString to pure str
        else:
            #print("%sChild %s" % (depth*". ",soup.name))
            subelement = elements_from_soup(child, remove_tags, keep_attributes)
            if subelement != None:
                element.append(subelement)
    return element

# abstract class


class HTMLElement(CompoundElement):

    """Abstract base class for all elements."""


class HTML(HTMLElement):

    """Element corresponding to the ``<html>`` tag"""


class Head(HTMLElement):

    """Element corresponding to the ``<head>`` tag"""
# a title cannot contain subelements -- derive from UnicodeElement instead?


class Title(HTMLElement):

    """Element corresponding to the ``<title>`` tag"""


class Body(HTMLElement):

    """Element corresponding to the ``<body>`` tag"""

    def as_xhtml(self, uri):
        element = super(Body, self).as_xhtml(uri)
        element.set('about', uri)
        return element

# %block


class P(HTMLElement):

    """Element corresponding to the ``<p>`` tag"""

# %heading


class H1(HTMLElement):

    """Element corresponding to the ``<h1>`` tag"""


class H2(HTMLElement):

    """Element corresponding to the ``<h2>`` tag"""


class H3(HTMLElement):

    """Element corresponding to the ``<h3>`` tag"""


class H4(HTMLElement):

    """Element corresponding to the ``<h4>`` tag"""


class H5(HTMLElement):

    """Element corresponding to the ``<h5>`` tag"""


class H6(HTMLElement):

    """Element corresponding to the ``<h6>`` tag"""

# %list


class UL(HTMLElement):

    """Element corresponding to the ``<ul>`` tag"""


class OL(HTMLElement):

    """Element corresponding to the ``<ol>`` tag"""


class LI(HTMLElement):

    """Element corresponding to the ``<li>`` tag"""

# %preformatted


class Pre(HTMLElement):

    """Element corresponding to the ``<pre>`` tag"""

# other


class DL(HTMLElement):

    """Element corresponding to the ``<dl>`` tag"""


class DT(HTMLElement):

    """Element corresponding to the ``<dt>`` tag"""


class DD(HTMLElement):

    """Element corresponding to the ``<dd>`` tag"""


class Div(HTMLElement):

    """Element corresponding to the ``<div>`` tag"""


class Blockquote(HTMLElement):

    """Element corresponding to the ``<blockquote>`` tag"""


class Form(HTMLElement):

    """Element corresponding to the ``<form>`` tag"""


class HR(HTMLElement):

    """Element corresponding to the ``<hr>`` tag"""


class Table(HTMLElement):

    """Element corresponding to the ``<table>`` tag"""


class Fieldset(HTMLElement):

    """Element corresponding to the ``<fieldset>`` tag"""


class Address(HTMLElement):

    """Element corresponding to the ``<address>`` tag"""

# %fontstyle


class TT (HTMLElement):

    """Element corresponding to the ``<tt >`` tag"""


class I (HTMLElement):

    """Element corresponding to the ``<i >`` tag"""


class B (HTMLElement):

    """Element corresponding to the ``<b >`` tag"""


class U (HTMLElement):

    """Element corresponding to the ``<u >`` tag"""


class Big (HTMLElement):

    """Element corresponding to the ``<big >`` tag"""


class Small(HTMLElement):

    """Element corresponding to the ``<small>`` tag"""

# %phrase


class Em (HTMLElement):

    """Element corresponding to the ``<em >`` tag"""


class Strong (HTMLElement):

    """Element corresponding to the ``<strong >`` tag"""


class Dfn (HTMLElement):

    """Element corresponding to the ``<dfn >`` tag"""


class Code (HTMLElement):

    """Element corresponding to the ``<code >`` tag"""


class Samp (HTMLElement):

    """Element corresponding to the ``<samp >`` tag"""


class Kbd (HTMLElement):

    """Element corresponding to the ``<kbd >`` tag"""


class Var (HTMLElement):

    """Element corresponding to the ``<var >`` tag"""


class Cite (HTMLElement):

    """Element corresponding to the ``<cite >`` tag"""


class Abbr (HTMLElement):

    """Element corresponding to the ``<abbr >`` tag"""


class Acronym(HTMLElement):

    """Element corresponding to the ``<acronym>`` tag"""

# %special


class A (HTMLElement):

    """Element corresponding to the ``<a >`` tag"""


class Img (HTMLElement):

    """Element corresponding to the ``<img >`` tag"""


class Object (HTMLElement):

    """Element corresponding to the ``<object >`` tag"""


class Br (HTMLElement):

    """Element corresponding to the ``<br >`` tag"""


class Q (HTMLElement):

    """Element corresponding to the ``<q >`` tag"""


class Sub (HTMLElement):

    """Element corresponding to the ``<sub >`` tag"""


class Sup (HTMLElement):

    """Element corresponding to the ``<sup >`` tag"""


class Span (HTMLElement):

    """Element corresponding to the ``<span >`` tag"""


class BDO(HTMLElement):

    """Element corresponding to the ``<bdo>`` tag"""

# %form


class Input(HTMLElement):

    """Element corresponding to the ``<input>`` tag"""


class Select(HTMLElement):

    """Element corresponding to the ``<select>`` tag"""


class Textarea(HTMLElement):

    """Element corresponding to the ``<textarea>`` tag"""


class Label(HTMLElement):

    """Element corresponding to the ``<label>`` tag"""


class Button(HTMLElement):

    """Element corresponding to the ``<button>`` tag"""

# table


class Caption(HTMLElement):

    """Element corresponding to the ``<caption>`` tag"""


class Thead(HTMLElement):

    """Element corresponding to the ``<thead>`` tag"""


class Tfoot(HTMLElement):

    """Element corresponding to the ``<tfoot>`` tag"""


class Tbody(HTMLElement):

    """Element corresponding to the ``<tbody>`` tag"""


class Colgroup(HTMLElement):

    """Element corresponding to the ``<colgroup>`` tag"""


class Col(HTMLElement):

    """Element corresponding to the ``<col>`` tag"""


class TR(HTMLElement):

    """Element corresponding to the ``<tr>`` tag"""


class TH(HTMLElement):

    """Element corresponding to the ``<th>`` tag"""


class TD(HTMLElement):

    """Element corresponding to the ``<td>`` tag"""

# very special?


class Ins(HTMLElement):

    """Element corresponding to the ``<ins>`` tag"""


class Del(HTMLElement):

    """Element corresponding to the ``<del>`` tag"""

# new elements in HTML5 -- cannot be simply expressed in XHTML
# 1.1. Instead they're expressed as eg. '<div class="section">'


class HTML5Element(HTMLElement):
    tagname = "div"

    def _get_classname(self):
        return self.__class__.__name__.lower()
    classname = property(_get_classname)


class Article(HTML5Element):

    """Element corresponding to the ``<article>`` tag"""


class Aside(HTML5Element):

    """Element corresponding to the ``<aside>`` tag"""


class Bdi(HTML5Element):

    """Element corresponding to the ``<bdi>`` tag"""


class Details(HTML5Element):

    """Element corresponding to the ``<details>`` tag"""


class Dialog(HTML5Element):

    """Element corresponding to the ``<dialog>`` tag"""


class Summary(HTML5Element):

    """Element corresponding to the ``<summary>`` tag"""


class Figure(HTML5Element):

    """Element corresponding to the ``<figure>`` tag"""


class Figcaption(HTML5Element):

    """Element corresponding to the ``<figcaption>`` tag"""


class Footer(HTML5Element):

    """Element corresponding to the ``<footer>`` tag"""


class Header(HTML5Element):

    """Element corresponding to the ``<header>`` tag"""


class Hgroup(HTML5Element):

    """Element corresponding to the ``<hgroup>`` tag"""


class Mark(HTML5Element):

    """Element corresponding to the ``<mark>`` tag"""


class Meter(HTML5Element):

    """Element corresponding to the ``<meter>`` tag"""


class Nav(HTML5Element):

    """Element corresponding to the ``<nav>`` tag"""


class Progress(HTML5Element):

    """Element corresponding to the ``<progress>`` tag"""


class Ruby(HTML5Element):

    """Element corresponding to the ``<ruby>`` tag"""


class Rt(HTML5Element):

    """Element corresponding to the ``<rt>`` tag"""


class Rp(HTML5Element):

    """Element corresponding to the ``<rp>`` tag"""


class Section(HTML5Element):

    """Element corresponding to the ``<section>`` tag"""


class Time(HTML5Element):

    """Element corresponding to the ``<time>`` tag"""


class Wbr(HTML5Element):

    """Element corresponding to the ``<wbr>`` tag"""
# audio, video, embed, canvas and similar non structural/semantic
# elements not included

# For use by elements_from_soup. FIXME: we should be able to build
# _tagmap dynamically.
_tagmap = {'html': HTML,
           'head': Head,
           'title': Title,
           'body': Body,
           'p': P,
           'h1': H1,
           'h2': H2,
           'h3': H3,
           'h4': H4,
           'h5': H5,
           'h6': H6,
           'ul': UL,
           'ol': OL,
           'li': LI,
           'pre': Pre,
           'dl': DL,
           'dt': DT,
           'dd': DD,
           'div': Div,
           'blockquote': Blockquote,
           'form': Form,
           'hr': HR,
           'table': Table,
           'fieldset': Fieldset,
           'address': Address,
           'tt': TT,
           'i': I,
           'b': B,
           'u': U,
           'big': Big,
           'small': Small,
           'em': Em,
           'strong': Strong,
           'dfn': Dfn,
           'code': Code,
           'samp': Samp,
           'kbd': Kbd,
           'var': Var,
           'cite': Cite,
           'abbr': Abbr,
           'acronym': Acronym,
           'a': A,
           'img': Img,
           'object': Object,
           'br': Br,
           'q': Q,
           'sub': Sub,
           'sup': Sup,
           'span': Span,
           'bdo': BDO,
           'input': Input,
           'select': Select,
           'textarea': Textarea,
           'label': Label,
           'button': Button,
           'caption': Caption,
           'thead': Thead,
           'tfoot': Tfoot,
           'tbody': Tbody,
           'colgroup': Colgroup,
           'col': Col,
           'tr': TR,
           'th': TH,
           'td': TD,
           'ins': Ins,
           'del': Del,
           'article': Article,
           'aside': Aside,
           'bdi': Bdi,
           'details': Details,
           'dialog': Dialog,
           'summary': Summary,
           'figure': Figure,
           'figcaption': Figcaption,
           'footer': Footer,
           'header': Header,
           'hgroup': Hgroup,
           'mark': Mark,
           'meter': Meter,
           'nav': Nav,
           'progress': Progress,
           'ruby': Ruby,
           'rt': Rt,
           'rp': Rp,
           'section': Section,
           'time': Time,
           'wbr': Wbr
           }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

from ferenda import util
from ferenda import DocumentRepository, Describer, PDFReader
from ferenda.decorators import managedparsing
from ferenda.elements import Body


class PDFDocumentRepository(DocumentRepository):

    """Base class for handling repositories of PDF documents. Parsing
    of these documents are a bit more complicated than HTML or text
    documents, particularly with the handling of external resources
    such as CSS and image files."""
    storage_policy = "dir"
    downloaded_suffix = ".pdf"

    def get_default_options(self):
        opts = super(PDFDocumentRepository, self).get_default_options()
        opts['pdfimages'] = True
        return opts

    @managedparsing
    def parse(self, doc):
        reader = self.pdfreader_from_basefile(doc.basefile)
        self.parse_from_pdfreader(reader, doc)
        # return doc
        return True

    def pdfreader_from_basefile(self, basefile):
        pdffile = self.store.downloaded_path(basefile)
        # Convoluted way of getting the directory of the intermediate
        # xml + png files that PDFReader will create

        intermediate_dir = os.path.dirname(self.store.intermediate_path(basefile))
        pdf = PDFReader()

        if self.config.compress == "bz2":
            keep_xml = "bz2"
        else:
            keep_xml = True
        pdf.read(pdffile, intermediate_dir,
                 images=self.config.pdfimages,
                 keep_xml=keep_xml)
        return pdf

    def parse_from_pdfreader(self, pdfreader, doc):
        doc.body = Body([pdfreader])

        d = Describer(doc.meta, doc.uri)
        d.rdftype(self.rdf_type)
        d.value(self.ns['prov'].wasGeneratedBy, self.qualified_class_name())

        return doc

    def create_external_resources(self, doc):
        cssfile = self.store.parsed_path(doc.basefile, attachment="index.css")
        with open(cssfile, "w") as fp:
            # Create CSS header with fontspecs
            for pdf in doc.body:
                assert isinstance(pdf, PDFReader)
                for spec in list(pdf.fontspec.values()):
                    fp.write(".fontspec%s {font: %spx %s; color: %s;}\n" %
                             (spec['id'], spec['size'], spec['family'], spec['color']))

            # 2 Copy all created png files to their correct locations
            totcnt = 0
            pdfbase = os.path.splitext(os.path.basename(pdf.filename))[0]
            for pdf in doc.body:
                cnt = 0
                for page in pdf:
                    totcnt += 1
                    cnt += 1
                    if page.background:
                        src = self.store.intermediate_path(
                            doc.basefile, attachment=os.path.basename(page.background))
                        dest = self.store.parsed_path(
                            doc.basefile, attachment=os.path.basename(page.background))
                        if util.copy_if_different(src, dest):
                            self.log.debug("Copied %s to %s" % (src, dest))

                        fp.write("#page%03d { background: url('%s');}\n" %
                                 (cnt, os.path.basename(dest)))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import hashlib
import json
from datetime import datetime


from ferenda import util


class DocumentEntry(object):

    """This class has two primary uses -- it is used to represent and store
    aspects of the downloading of each document (when it was initially
    downloaded, optionally updated, and last checked, as well as the URL
    from which it was downloaded). It's also used by the news_* methods
    to encapsulate various aspects of a document entry in an atom
    feed. Some properties and methods are used by both of these use
    cases, but not all.

    :param path: If this file path is an existing JSON file, the object is initialized from that file.
    :type  path: str
    """
    id = None
    """The canonical uri for the document."""

    basefile = None
    """The basefile for the document."""

    orig_created = None
    """The first time we fetched the document from it's original location."""

    orig_updated = None
    """The last time the content at the original location of the
    document was changed."""

    orig_checked = None
    """The last time we accessed the original location of this
    document, regardless of wheter this led to an update."""

    orig_url = None
    """The main url from where we fetched this document."""

    published = None
    """The date our parsed/processed version of the document was published."""

    updated = None
    """The last time our parsed/processed version changed in any way
       (due to the original content being updated, or due to changes
       in our parsing functionality."""

    url = None
    """The URL to the browser-ready version of the page, equivalent to what
    :meth:`~ferenda.DocumentRepository.generated_url` returns."""

    title = None
    """A title/label for the document, as used in an Atom feed."""

    summary = None
    """A summary of the document, as used in an Atom feed."""

    content = None
    """A dict that represents metadata about the document file."""

    link = None
    """A dict that represents metadata about the document RDF metadata
    (such as it's URI, length, MIME-type and MD5 hash)."""

    def __init__(self, path=None):
        # for json serialization
        def myhook(d):
            for key in ('orig_created', 'orig_updated', 'orig_checked', 'published', 'updated'):
                if key in d and d[key]:
                    try:
                        dt = datetime.strptime(d[key], '%Y-%m-%dT%H:%M:%S.%f')
                    except ValueError:
                        # no fractional part
                        dt = datetime.strptime(d[key], '%Y-%m-%dT%H:%M:%S')
                    d[key] = dt
            return d

        if path and os.path.exists(path):
            with open(path) as fp:
                d = json.load(fp, object_hook=myhook)
            self.__dict__.update(d)
            self._path = path
        else:
            self.id = None
            self.basefile = None
            self.orig_updated = None
            self.orig_checked = None
            self.orig_url = None
            self.published = None
            self.updated = None
            self.title = None
            self.summary = None
            self.url = None
            if path:
                self._path = path
            # Content src="...": A link to the actual document, or the
            # content inline (Source or refined version?)
            self.content = {}
            # Link rel="alternate": The metadata for this document (and
            # included resources)
            self.link = {}


    def __repr__(self):
        return '<%s id=%s>' % (self.__class__.__name__, self.id)

    def save(self, path=None):
        """Saves the state of the documententry to a JSON file at *path*. If
*path* is not provided, uses the path that the object was initialized
with.

        """

        def mydefault(obj):
            if isinstance(obj, datetime):
                return obj.isoformat()
            raise TypeError("%r is not JSON serializable" % obj)

        if not path:
            path = self._path  # better be there
        d = dict((k, v) for (k, v) in self.__dict__.items() if k[0] != "_")
        util.ensure_dir(path)
        with open(path, "w") as fp:
            json.dump(d, fp, default=mydefault, indent=2, separators=(', ',': '), sort_keys=True)
    # If inline=True, the contents of filename is included in the Atom
    # entry. Otherwise, it just references it.
    #
    # Note that you can only have one content element.

    def set_content(self, filename, url, mimetype=None, inline=False):
        """Sets the ``content`` property and calculates md5 hash for the file

        :param filename: The full path to the document file
        :param url: The full external URL that will be used to get the same document file
        :param mimetype: The MIME-type used in the atom feed. If not provided, guess from file extension.
        :param inline: whether to inline the document content in the file or refer to *url*
        """
        if not mimetype:
            mimetype = self.guess_type(filename)
        self.content['type'] = mimetype
        if inline:
            # there's a difference between actual mimetype and
            # mimetype-as-type-in-atom.
            if mimetype == "application/html+xml":
                mimetype = "xhtml"
            assert mimetype == 'xhtml', "Can't inline non-xhtml content"
            with open(filename) as fp:
                self.content['markup'] = fp.read()
            self.content['src'] = None
            self.content['hash'] = None
        else:
            self.content['markup'] = None
            self.content['src'] = url
            self.content['hash'] = "md5:%s" % self.calculate_md5(filename)

    def set_link(self, filename, url, mimetype=None):
        """Sets the ``link`` property and calculate md5 hash for the RDF metadata.

        :param filename: The full path to the RDF file for a document
        :param url: The full external URL that will be used to get the same RDF file
        :param mimetype: The MIME-type used in the atom feed. If not provided, guess from file extension.
        """
        if not mimetype:
            mimetype = self.guess_type(filename)
        self.link['href'] = url
        self.link['type'] = mimetype
        self.link['length'] = os.path.getsize(filename)
        self.link['hash'] = "md5:%s" % self.calculate_md5(filename)

    def calculate_md5(self, filename):
        """Given a filename, return the md5 value for the file's content."""
        c = hashlib.md5()
        with open(filename, 'rb') as fp:
            c.update(fp.read())
        return c.hexdigest()

    def guess_type(self, filename):
        """Given a filename, return a MIME-type based on the file extension."""
        exts = {'.pdf': 'application/pdf',
                '.rdf': 'application/rdf+xml',
                '.html': 'text/html',
                '.xhtml': 'application/html+xml'}
        for ext, mimetype in list(exts.items()):
            if filename.endswith(ext):
                return mimetype
        return "application/octet-stream"



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


class TocPage(object):

    """Represents a particular TOC page.
    
    :param linktext: The text used for TOC links *to* this page, like "a" or "2013".
    :param linktext: str
    :param label: A description of this page, like "Documents starting with 'a'"
    :type  label: str
    :param binding: The variable binding used for defining this TOC page, like "title" or "issued"
    :type  binding: str
    :param value: The particular value of bound variable that corresponds to this TOC page, like "a" or "2013". The ``selector`` function of a :py:class:`~ferenda.TocCriteria` object is used to select this value out of the raw data.
    :type  value: str
    """

    def __init__(self, linktext, title, binding, value):
        self.linktext = linktext
        self.title = title
        self.binding = binding
        self.value = value

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        dictrepr = "".join((" %s=%s" % (k, v) for k, v in sorted(self.__dict__.items())))
        return ("<%s%s>" % (self.__class__.__name__, dictrepr))
